#  Licensed to the Apache Software Foundation (ASF) under one or more
#  contributor license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright ownership.
#  The ASF licenses this file to You under the Apache License, Version 2.0
#  (the "License"); you may not use this file except in compliance with
#  the License. You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import copy
import json

from typing import Any, Dict, Iterable, List, Optional, Union

from pydantic import BaseModel

from metadata.generated.schema.type.entityReference import EntityReference
from metadata.ingestion.models.json_serializable import JsonSerializable

DESCRIPTION_NODE_LABEL_VAL = 'Description'
DESCRIPTION_NODE_LABEL = DESCRIPTION_NODE_LABEL_VAL


class ColumnMetadata:

    def __init__(self,
                 name: str,
                 documentation: Union[str, None],
                 column_data_type: str,
                 ordinal_position: int,
                 pii_tags: List[str] = None
                 ) -> None:
        """
        TODO: Add stats
        :param name:
        :param documentation:
        :param column_data_type:
        :param ordinal_position:
        :param pii_tags:
        """
        self.name = name
        self.documentation = documentation
        self.column_data_type = column_data_type
        self.ordinal_position = ordinal_position
        self.pii_tags = pii_tags if pii_tags else []

    def __repr__(self) -> str:
        return 'ColumnMetadata({!r}, {!r}, {!r}, {!r}, {!r})'.format(self.name,
                                                                     self.documentation,
                                                                     self.column_data_type,
                                                                     self.ordinal_position,
                                                                     self.pii_tags)


class TableMetadata:
    """
    Table metadata that contains columns.
    This class can be used for both table and view metadata. If it is a View, is_view=True should be passed in.
    """

    def __init__(self,
                 name: str,
                 documentation: Union[str, None],
                 columns: Iterable[ColumnMetadata] = None
                 ) -> None:
        """
        :param name:
        :param documentation:
        :param columns:
        """
        self.name = name
        self.documentation = documentation
        self.columns = columns if columns else []

    def add_column(self, column: ColumnMetadata):
        self.columns.append(column)

    def to_json(self):
        return json.dumps(self, default=lambda o: o.__dict__)

    def __repr__(self) -> str:
        return 'TableMetadata({!r}, {!r}, {!r} '.format(self.name,
                                                        self.documentation,
                                                        self.columns)


class DatabaseMetadata(JsonSerializable):
    """
    Database metadata that contains Tables.
    This class can be used to generate catalog batch REST api representation
    All connectors should use this to build JSON representation of catalog REST API.
    """
    DATABASE_NODE_LABEL = 'database'
    CLUSTER_NODE_LABEL = 'Cluster'

    @classmethod
    def create(cls, record: Dict[str, Any], service: Dict[str, Any]):
        name = record.name
        documentation = record.documentation
        service = service
        tables = []
        for table in record.tables:
            table_metadata: TableMetadata = TableMetadata(table.name, table.documentation)
            for column in table.columns:
                column: ColumnMetadata = ColumnMetadata(column.name, column.documentation,
                                                        column.column_data_type, column.ordinal_position,
                                                        column.pii_tags)
                table_metadata.add_column(column)
            tables.append(table_metadata)
        return cls(name, documentation, tables, service)

    def __init__(self,
                 name: str,
                 documentation: Union[str, None],
                 tables: Iterable[TableMetadata] = None,
                 service: Dict[str, Any] = None,
                 **kwargs: Any
                 ) -> None:
        """
        :param name:
        :param service:
        :param documentation:
        :param tables:
        :param kwargs: Put additional attributes to the table model if there is any.
        """
        self.name = name
        self.documentation = documentation
        self.tables = tables if tables else []
        self.service = service
        if kwargs:
            self.attrs = copy.deepcopy(kwargs)

    def __repr__(self) -> str:
        return 'DatabaseMetadata({!r}, {!r}, {!r})'.format(self.name,
                                                           self.documentation,
                                                           self.tables)

    def add_table(self, table: TableMetadata) -> None:
        self.tables.append(table)


class JDBCMetadata(JsonSerializable):
    def __init__(self,
                 connection_url: str,
                 driver_class: str):
        self.connection_url = connection_url
        self.driver_class = driver_class


class ServiceMetadata(JsonSerializable):
    """
    Service Metadata contains the configuration to connect to a database
    """

    def __init__(self,
                 name: str,
                 jdbc: JDBCMetadata):
        self.name = name
        self.jdbc = jdbc


class TableESDocument(BaseModel):
    """ Elastic Search Mapping doc """
    table_id: str
    database: str
    service: str
    service_type: str
    table_name: str
    suggest: List[dict]
    description: Optional[str] = None
    table_type: Optional[str] = None
    last_updated_timestamp: Optional[int]
    column_names: List[str]
    column_descriptions: List[str]
    monthly_stats: int
    monthly_percentile_rank: int
    weekly_stats: int
    weekly_percentile_rank: int
    daily_stats: int
    daily_percentile_rank: int
    tags: List[str]
    fqdn: str
    tier: Optional[str] = None
    schema_description: Optional[str] = None
    owner: str
    followers: List[str]


class TopicESDocument(BaseModel):
    """ Topic Elastic Search Mapping doc """
    topic_id: str
    service: str
    service_type: str
    topic_name: str
    suggest: List[dict]
    description: Optional[str] = None
    last_updated_timestamp: Optional[int]
    tags: List[str]
    fqdn: str
    tier: Optional[str] = None
    schema_description: Optional[str] = None
    owner: str
    followers: List[str]


class DashboardESDocument(BaseModel):
    """ Elastic Search Mapping doc for Dashboards """
    dashboard_id: str
    service: str
    service_type: str
    dashboard_name: str
    suggest: List[dict]
    description: Optional[str] = None
    last_updated_timestamp: Optional[int]
    chart_names: List[str]
    chart_descriptions: List[str]
    tags: List[str]
    fqdn: str
    tier: Optional[str] = None
    owner: str
    followers: List[str]
    monthly_stats: int
    monthly_percentile_rank: int
    weekly_stats: int
    weekly_percentile_rank: int
    daily_stats: int
    daily_percentile_rank: int


class DashboardOwner(BaseModel):
    """Dashboard owner"""
    username: str
    first_name: str
    last_name: str


class Chart(BaseModel):
    """Chart"""
    name: str
    displayName: str
    description: str
    chart_type: str
    url: str
    owners: List[DashboardOwner] = None
    lastModified: int = None
    datasource_fqn: str = None
    service: EntityReference
    custom_props: Dict[Any, Any] = None


class Dashboard(BaseModel):
    """Dashboard"""
    name: str
    displayName: str
    description: str
    url: str
    owners: List[DashboardOwner] = None
    charts: List[str]
    service: EntityReference
    lastModified: int = None


class ValueFrequency(BaseModel):
    """Profiler ValueFrequency"""
    value: str
    frequency: int


class Histogram(BaseModel):
    """Histogram"""
    boundaries: List[str]
    heights: List[str]


class Quantile(BaseModel):
    """Quantile"""
    quantile: str
    value: str


class DatasetColumnProfile(BaseModel):
    """Dataset Column Profile stats """
    fqdn: str
    unique_count: int = None
    unique_proportion: int = None
    null_count: int = None
    null_proportion: int = None
    min: str = None
    max: str = None
    mean: str = None
    median: str = None
    stddev: str = None
    quantiles: List[Quantile] = None
    distinct_value_frequencies: List[ValueFrequency] = None
    histogram: List[Histogram] = None
    sample_values: List[str] = None


class DatasetProfile(BaseModel):
    """Dataset(table) stats"""
    timestamp: int
    table_name: str
    row_count: int = None
    col_count: int = None
    col_profiles: List[DatasetColumnProfile] = None





#  Licensed to the Apache Software Foundation (ASF) under one or more
#  contributor license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright ownership.
#  The ASF licenses this file to You under the Apache License, Version 2.0
#  (the "License"); you may not use this file except in compliance with
#  the License. You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import uuid
from dataclasses import dataclass, field
from typing import Iterable, Optional, List

from metadata.ingestion.api.common import WorkflowContext
from metadata.ingestion.api.source import Source, SourceStatus
from metadata.ingestion.models.ometa_table_db import OMetaDatabaseAndTable
from simple_salesforce import Salesforce

from .sql_source import SQLConnectionConfig
from ..ometa.openmetadata_rest import MetadataServerConfig
from ...generated.schema.entity.data.database import Database
from ...generated.schema.entity.data.table import Column, ColumnConstraint, Table, TableData
from ...generated.schema.type.entityReference import EntityReference
from metadata.utils.helpers import get_database_service_or_create
from pydantic import ValidationError

logger: logging.Logger = logging.getLogger(__name__)

@dataclass
class SalesforceSourceStatus(SourceStatus):
    success: List[str] = field(default_factory=list)
    failures: List[str] = field(default_factory=list)
    warnings: List[str] = field(default_factory=list)
    filtered: List[str] = field(default_factory=list)

    def scanned(self, table_name: str) -> None:
        self.success.append(table_name)
        logger.info('Table Scanned: {}'.format(table_name))

    def filter(
            self, table_name: str, err: str, dataset_name: str = None, col_type: str = None
    ) -> None:
        self.filtered.append(table_name)
        logger.warning("Dropped Table {} due to {}".format(table_name, err))


class SalesforceConfig(SQLConnectionConfig):
    username: str
    password: str
    security_token: str
    host_port: Optional[str]
    scheme: str
    service_type = "MySQL"
    sobject_name: str

    def get_connection_url(self):
        return super().get_connection_url()


class SalesforceSource(Source):
    def __init__(self, config: SalesforceConfig, metadata_config: MetadataServerConfig, ctx):
        super().__init__(ctx)
        self.config = config
        self.service = get_database_service_or_create(config, metadata_config)
        self.status = SalesforceSourceStatus()
        self.sf = Salesforce(
                username=self.config.username, password=self.config.password,
                security_token=self.config.security_token
            )

    @classmethod
    def create(cls, config: dict, metadata_config: dict, ctx: WorkflowContext):
        config = SalesforceConfig.parse_obj(config)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config)
        return cls(config, metadata_config, ctx)

    def column_type(self, column_type: str):
        if column_type in ["ID", "PHONE", "CURRENCY"]:
            type = "INT"
        elif column_type in ["REFERENCE", "PICKLIST", "TEXTAREA", "ADDRESS", "URL"]:
            type = "VARCHAR"
        else:
            type = column_type
        return type

    def next_record(self) -> Iterable[OMetaDatabaseAndTable]:
        yield from self.salesforce_client()

    def fetch_sample_data(self,sobject_name):
        md = self.sf.restful("sobjects/{}/describe/".format(sobject_name), params=None)
        columns = []
        rows = []
        for column in md['fields']:
            columns.append(column['name'])
        query = "select {} from {}".format(str(columns)[1:-1].replace('\'',''),sobject_name)
        logger.info("Ingesting data using {}".format(query))
        resp = self.sf.query(query)
        for record in resp['records']:
            row = []
            for column in columns:
                row.append(record[f'{column}'])
            rows.append(row)
        return TableData(columns=columns, rows=rows)

    def salesforce_client(self) -> Iterable[OMetaDatabaseAndTable]:
        try:
            
            row_order = 1
            table_columns = []
            md = self.sf.restful("sobjects/{}/describe/".format(self.config.sobject_name), params=None)
            
            for column in md['fields']:
                col_constraint = None
                if column['nillable']:
                    col_constraint = ColumnConstraint.NULL
                elif not column['nillable']:
                    col_constraint = ColumnConstraint.NOT_NULL
                if column['unique']:
                    col_constraint = ColumnConstraint.UNIQUE

                table_columns.append(
                    Column(
                        name=column['name'],
                        description=column['label'],
                        columnDataType=self.column_type(column['type'].upper()),
                        columnConstraint=col_constraint,
                        ordinalPosition=row_order
                    )
                )
                row_order += 1
            table_data = self.fetch_sample_data(self.config.sobject_name)
            logger.info("Successfully Ingested the sample data")
            table_entity = Table(
                id=uuid.uuid4(),
                name=self.config.sobject_name,
                tableType='Regular',
                description=" ",
                columns=table_columns,
                sampleData=table_data
            )
            self.status.scanned(f"{self.config.scheme}.{self.config.sobject_name}")
            database_entity = Database(
                name=self.config.scheme,
                service=EntityReference(id=self.service.id, type=self.config.service_type)
            )
            table_and_db = OMetaDatabaseAndTable(table=table_entity, database=database_entity)
            yield table_and_db
        except ValidationError as err:
            logger.error(err)
            self.status.failure('{}'.format(self.config.sobject_name), err)

    def prepare(self):
        pass

    def get_status(self) -> SourceStatus:
        return self.status



#  Licensed to the Apache Software Foundation (ASF) under one or more
#  contributor license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright ownership.
#  The ASF licenses this file to You under the Apache License, Version 2.0
#  (the "License"); you may not use this file except in compliance with
#  the License. You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging

from metadata.config.common import ConfigModel
from metadata.ingestion.api.common import WorkflowContext, Record
from metadata.ingestion.api.sink import Sink, SinkStatus
from metadata.ingestion.models.user import MetadataTeam, MetadataUser
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig, OpenMetadataAPIClient
from metadata.ingestion.ometa.client import APIError

logger = logging.getLogger(__name__)


class MetadataUsersSinkConfig(ConfigModel):
    api_end_point: str = None


class MetadataRestUsersSink(Sink):
    config: MetadataUsersSinkConfig
    metadata_config: MetadataServerConfig
    status: SinkStatus

    def __init__(self, ctx: WorkflowContext, config: MetadataUsersSinkConfig, metadata_config: MetadataServerConfig):
        super().__init__(ctx)
        self.config = config
        self.metadata_config = metadata_config
        self.status = SinkStatus()
        self.api_team_post = "/teams"
        self.api_team_get = "/teams"
        self.api_users = "/users"
        self.org_entities = {}
        self.role_entities = {}
        self.team_entities = {}
        self.client = OpenMetadataAPIClient(self.metadata_config).client
        self._bootstrap_entities()

    @classmethod
    def create(cls, config_dict: dict, metadata_config_dict: dict, ctx: WorkflowContext):
        config = MetadataUsersSinkConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(ctx, config, metadata_config)

    def write_record(self, record: Record) -> None:
        self._create_user(record)

    def _bootstrap_entities(self):
        # Fetch teams per org

        team_response = self.client.get(self.api_team_get)
        for team in team_response['data']:
            self.team_entities[team['displayName']] = team['id']

    def _create_team(self, record: MetadataUser) -> None:
        team_name = record.team_name
        metadata_team = MetadataTeam(team_name, 'Team Name')
        try:
            r = self.client.post(self.api_team_post,
                                 data=metadata_team.to_json()
                                 )
            instance_id = r['id']
            self.team_entities[team_name] = instance_id
        except APIError:
            pass

    def _create_user(self, record: MetadataUser) -> None:
        if record.team_name not in self.team_entities:
            self._create_team(record)
        teams = [self.team_entities[record.team_name]]
        # Using github username for generating a login name
        metadata_user = MetadataUser(name=record.github_username,
                                     display_name=record.name,
                                     email=record.email,
                                     teams=teams)
        try:
            self.client.post(self.api_users, data=metadata_user.to_json())
            self.status.records_written(record.github_username)
            logger.info("Sink: {}".format(record.github_username))
        except APIError:
            pass

    def get_status(self) -> SinkStatus:
        return self.status

    def close(self):
        pass



# generated by datamodel-codegen:
#   filename:  schema/entity/data/chart.json
#   timestamp: 2021-09-13T04:07:21+00:00

from __future__ import annotations

from enum import Enum
from typing import List, Optional

from pydantic import AnyUrl, BaseModel, Field, constr

from ...type import basic, entityReference, tagLabel, usageDetails


class ChartType(Enum):
    Line = 'Line'
    Table = 'Table'
    Bar = 'Bar'
    Area = 'Area'
    Pie = 'Pie'
    Histogram = 'Histogram'
    Scatter = 'Scatter'
    Text = 'Text'
    BoxPlot = 'BoxPlot'
    Other = 'Other'


class Chart(BaseModel):
    id: basic.Uuid = Field(
        ..., description='Unique identifier that identifies a chart instance.'
    )
    name: constr(min_length=1, max_length=64) = Field(
        ..., description='Name that identifies this Chart.'
    )
    displayName: Optional[str] = Field(
        None,
        description='Display Name that identifies this Chart. It could be title or label from the source services',
    )
    fullyQualifiedName: Optional[constr(min_length=1, max_length=64)] = Field(
        None,
        description="A unique name that identifies a dashboard in the format 'ServiceName.ChartName'.",
    )
    description: Optional[str] = Field(
        None, description='Description of the dashboard, what it is, and how to use it.'
    )
    chartType: Optional[ChartType] = None
    chartUrl: Optional[AnyUrl] = Field(
        None, description='Chart URL, pointing to its own Service URL'
    )
    href: Optional[basic.Href] = Field(
        None, description='Link to the resource corresponding to this entity.'
    )
    owner: Optional[entityReference.EntityReference] = Field(
        None, description='Owner of this dashboard.'
    )
    tables: Optional[entityReference.EntityReferenceList] = Field(
        None, description='Link to table used in this chart.'
    )
    followers: Optional[entityReference.EntityReferenceList] = Field(
        None, description='Followers of this chart.'
    )
    tags: Optional[List[tagLabel.TagLabel]] = Field(
        None, description='Tags for this chart.'
    )
    service: entityReference.EntityReference = Field(
        ..., description='Link to service where this dashboard is hosted in.'
    )
    usageSummary: Optional[usageDetails.TypeUsedToReturnUsageDetailsOfAnEntity] = Field(
        None, description='Latest usage information for this database.'
    )



#  Licensed to the Apache Software Foundation (ASF) under one or more
#  contributor license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright ownership.
#  The ASF licenses this file to You under the Apache License, Version 2.0
#  (the "License"); you may not use this file except in compliance with
#  the License. You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import importlib
import logging
import uuid
from typing import Optional, Type, TypeVar

import click
from pydantic import Field

from metadata.config.common import (
    ConfigModel,
    DynamicTypedConfig,
    WorkflowExecutionError,
)
from metadata.ingestion.api.bulk_sink import BulkSink
from metadata.ingestion.api.common import WorkflowContext
from metadata.ingestion.api.processor import Processor
from metadata.ingestion.api.sink import Sink
from metadata.ingestion.api.source import Source
from metadata.ingestion.api.stage import Stage
from metadata.ingestion.api.source import Source

logger = logging.getLogger(__name__)

T = TypeVar("T")


class WorkflowConfig(ConfigModel):
    run_id: str = Field(default_factory=lambda: str(uuid.uuid1()))
    source: DynamicTypedConfig
    metadata_server: DynamicTypedConfig
    processor: Optional[DynamicTypedConfig] = None
    sink: Optional[DynamicTypedConfig] = None
    stage: Optional[DynamicTypedConfig] = None
    bulk_sink: Optional[DynamicTypedConfig] = None


class Workflow:
    config: WorkflowConfig
    ctx: WorkflowContext
    source: Source
    processor: Processor
    stage: Stage
    sink: Sink
    bulk_sink: BulkSink
    report = {}

    def __init__(self, config: WorkflowConfig):
        self.config = config
        self.ctx = WorkflowContext(workflow_id=self.config.run_id)
        source_type = self.config.source.type
        source_class = self.get('metadata.ingestion.source.{}.{}Source'.format(
            self.typeClassFetch(source_type, True), self.typeClassFetch(source_type, False)))
        metadata_config = self.config.metadata_server.dict().get("config", {})
        self.source: Source = source_class.create(
            self.config.source.dict().get("config", {}), metadata_config, self.ctx
        )
        logger.debug(f"Source type:{source_type},{source_class} configured")
        self.source.prepare()
        logger.debug(f"Source type:{source_type},{source_class}  prepared")

        if self.config.processor:
            processor_type = self.config.processor.type
            processor_class = self.get('metadata.ingestion.processor.{}.{}Processor'.format(
                self.typeClassFetch(processor_type, True), self.typeClassFetch(processor_type, False)))
            processor_config = self.config.processor.dict().get("config", {})
            self.processor: Processor = processor_class.create(processor_config, metadata_config, self.ctx)
            logger.debug(f"Processor Type: {processor_type}, {processor_class} configured")

        if self.config.stage:
            stage_type = self.config.stage.type
            stage_class = self.get('metadata.ingestion.stage.{}.{}Stage'.format(
                self.typeClassFetch(stage_type, True), self.typeClassFetch(stage_type, False)))
            stage_config = self.config.stage.dict().get("config", {})
            self.stage: Stage = stage_class.create(stage_config, metadata_config, self.ctx)
            logger.debug(f"Stage Type: {stage_type}, {stage_class} configured")

        if self.config.sink:
            sink_type = self.config.sink.type
            sink_class = self.get('metadata.ingestion.sink.{}.{}Sink'.format(
                self.typeClassFetch(sink_type, True), self.typeClassFetch(sink_type, False)))
            sink_config = self.config.sink.dict().get("config", {})
            self.sink: Sink = sink_class.create(sink_config, metadata_config, self.ctx)
            logger.debug(f"Sink type:{self.config.sink.type},{sink_class} configured")

        if self.config.bulk_sink:
            bulk_sink_type = self.config.bulk_sink.type
            bulk_sink_class = self.get('metadata.ingestion.bulksink.{}.{}BulkSink'.format(
                self.typeClassFetch(bulk_sink_type, True), self.typeClassFetch(bulk_sink_type, False)))
            bulk_sink_config = self.config.bulk_sink.dict().get("config", {})
            self.bulk_sink: BulkSink = bulk_sink_class.create(bulk_sink_config, metadata_config, self.ctx)
            logger.info(f"BulkSink type:{self.config.bulk_sink.type},{bulk_sink_class} configured")

    def typeClassFetch(self, type: str, isFile: bool):
        if isFile:
            return type.replace('-', '_')
        else:
            return ''.join([i.title() for i in type.replace('-', '_').split('_')])

    def get(self, key: str) -> Type[T]:
        if key.find(".") >= 0:
            # If the key contains a dot, we treat it as a import path and attempt
            # to load it dynamically.
            module_name, class_name = key.rsplit(".", 1)
            MyClass = getattr(importlib.import_module(module_name), class_name)
            return MyClass

    @classmethod
    def create(cls, config_dict: dict) -> "Workflow":
        config = WorkflowConfig.parse_obj(config_dict)
        return cls(config)

    def execute(self):
        for record in self.source.next_record():
            self.report['Source'] = self.source.get_status().as_obj()
            if hasattr(self, 'processor'):
                processed_record = self.processor.process(record)
            else:
                processed_record = record
            if hasattr(self, 'stage'):
                self.stage.stage_record(processed_record)
                self.report['Stage'] = self.stage.get_status().as_obj()
            if hasattr(self, 'sink'):
                self.sink.write_record(processed_record)
                self.report['sink'] = self.sink.get_status().as_obj()

    def stop(self):
        if hasattr(self, 'processor'):
            self.processor.close()
        if hasattr(self, 'stage'):
            self.stage.close()
        if hasattr(self, 'bulk_sink'):
            self.bulk_sink.write_records()
            self.report['Bulk_Sink'] = self.bulk_sink.get_status().as_obj()
            self.bulk_sink.close()
        if hasattr(self, 'sink'):
            self.sink.close()
        self.source.close()

    def raise_from_status(self, raise_warnings=False):
        if self.source.get_status().failures:
            raise WorkflowExecutionError(
                "Source reported errors", self.source.get_status()
            )
        if hasattr(self, 'sink') and self.sink.get_status().failures:
            raise WorkflowExecutionError("Sink reported errors", self.sink.get_status())
        if hasattr(self, 'stage') and self.stage.get_status().failures:
            raise WorkflowExecutionError("stage reported errors", self.stage.get_status())
        if hasattr(self, 'bulk_sink') and self.bulk_sink.get_status().failures:
            raise WorkflowExecutionError("Bulk Sink reported errors", self.bulk_sink.get_status())
        if raise_warnings and (
                self.source.get_status().warnings or self.sink.get_status().warnings
        ):
            raise WorkflowExecutionError(
                "Source reported warnings", self.source.get_status()
            )

    def print_status(self) -> int:
        click.echo()
        click.secho("Source Status:", bold=True)
        click.echo(self.source.get_status().as_string())
        if hasattr(self, 'stage'):
            click.secho("Stage Status:", bold=True)
            click.echo(self.stage.get_status().as_string())
            click.echo()
        if hasattr(self, 'sink'):
            click.secho("Sink Status:", bold=True)
            click.echo(self.sink.get_status().as_string())
            click.echo()
        if hasattr(self, 'bulk_sink'):
            click.secho("Bulk Sink Status:", bold=True)
            click.echo(self.bulk_sink.get_status().as_string())
            click.echo()

        if self.source.get_status().failures or (hasattr(self, 'sink') and self.sink.get_status().failures):
            click.secho("Workflow finished with failures", fg="bright_red", bold=True)
            return 1
        elif self.source.get_status().warnings or (hasattr(self, 'sink') and self.sink.get_status().warnings):
            click.secho("Workflow finished with warnings", fg="yellow", bold=True)
            return 0
        else:
            click.secho("Workflow finished successfully", fg="green", bold=True)
            return 0



# generated by datamodel-codegen:
#   filename:  schema/entity/data/topic.json
#   timestamp: 2021-09-13T04:07:21+00:00

from __future__ import annotations

from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, Field, conint, constr

from ...type import basic, entityReference, tagLabel


class TopicName(BaseModel):
    __root__: constr(regex=r'^[^.]*$', min_length=1, max_length=64) = Field(
        ..., description='Name that identifies a topic.'
    )


class SchemaType(Enum):
    Avro = 'Avro'
    Protobuf = 'Protobuf'
    JSON = 'JSON'
    Other = 'Other'


class CleanupPolicy(Enum):
    delete = 'delete'
    compact = 'compact'


class Topic(BaseModel):
    id: basic.Uuid = Field(
        ..., description='Unique identifier that identifies this topic instance.'
    )
    name: TopicName = Field(..., description='Name that identifies the topic.')
    fullyQualifiedName: Optional[str] = Field(
        None,
        description="Name that uniquely identifies a topic in the format 'messagingServiceName.topicName'.",
    )
    description: Optional[str] = Field(
        None, description='Description of the topic instance.'
    )
    service: entityReference.EntityReference = Field(
        ...,
        description='Link to the messaging cluster/service where this topic is hosted in.',
    )
    partitions: conint(ge=1) = Field(
        ..., description='Number of partitions into which the topic is divided.'
    )
    schemaText: Optional[str] = Field(
        None,
        description='Schema used for message serialization. Optional as some topics may not have associated schemas.',
    )
    schemaType: Optional[SchemaType] = Field(
        None, description='Schema used for message serialization.'
    )
    cleanupPolicies: Optional[List[CleanupPolicy]] = Field(
        None,
        description='Topic clean up policies. For Kafka - `cleanup.policy` configuration.',
    )
    retentionTime: Optional[float] = Field(
        None,
        description='Retention time in milliseconds. For Kafka - `retention.ms` configuration.',
    )
    replicationFactor: Optional[int] = Field(
        None, description='Replication Factor in integer (more than 1).'
    )
    maximumMessageSize: Optional[int] = Field(
        None,
        description='Maximum message size in bytes. For Kafka - `max.message.bytes` configuration.',
    )
    minimumInSyncReplicas: Optional[int] = Field(
        None,
        description='Minimum number replicas in sync to control durability. For Kafka - `min.insync.replicas` configuration.',
    )
    retentionSize: Optional[float] = Field(
        '-1',
        description='Maximum size of a partition in bytes before old data is discarded. For Kafka - `retention.bytes` configuration.',
    )
    owner: Optional[entityReference.EntityReference] = Field(
        None, description='Owner of this topic.'
    )
    followers: Optional[entityReference.EntityReferenceList] = Field(
        None, description='Followers of this table.'
    )
    tags: Optional[List[tagLabel.TagLabel]] = Field(
        None, description='Tags for this table.'
    )
    href: Optional[basic.Href] = Field(
        None, description='Link to the resource corresponding to this entity.'
    )



#  Licensed to the Apache Software Foundation (ASF) under one or more
#  contributor license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright ownership.
#  The ASF licenses this file to You under the Apache License, Version 2.0
#  (the "License"); you may not use this file except in compliance with
#  the License. You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict, Set
import os

from setuptools import setup, find_namespace_packages


def get_version():
    root = os.path.dirname(__file__)
    changelog = os.path.join(root, "CHANGELOG")
    with open(changelog) as f:
        return f.readline().strip()


def get_long_description():
    root = os.path.dirname(__file__)
    with open(os.path.join(root, "README.md")) as f:
        description = f.read()
    description += "\n\nChangelog\n=========\n\n"
    with open(os.path.join(root, "CHANGELOG")) as f:
        description += f.read()
    return description


scheduler_requirements = {
    "apns",
    "openmetadata-simplescheduler"
}


base_requirements = {
    "commonregex",
    "idna<3,>=2.5",
    "click<7.2.0,>=7.1.1",
    "expandvars>=0.6.5"
    "dataclasses>=0.8"
    "typing_extensions>=3.7.4"
    "mypy_extensions>=0.4.3",
    "typing-inspect",
    "pydantic==1.7.4",
    "pydantic[email]>=1.7.2",
    "google>=3.0.0",
    "google-auth>=1.33.0",
    "python-dateutil>=2.8.1",
    "email-validator>=1.0.3",
    "wheel~=0.36.2",
    "python-jose==3.3.0",
    "okta>=1.7.0",
    "sqlalchemy>=1.3.24",
    "sql-metadata~=2.0.0",
    "requests~=2.25.1"
}
pii_requirements = {
    "pandas~=1.3.1",
    "spacy==3.0.5"
}

base_plugins = {
    "query-parser",
    "metadata-usage",
    "file-stage",
    "sql-metadata~=2.0.0",
}
plugins: Dict[str, Set[str]] = {
    "athena": {"PyAthena[SQLAlchemy]"},
    "bigquery": {"pybigquery >= 0.6.0"},
    "bigquery-usage": {"google-cloud-logging", "cachetools"},
    "elasticsearch": {"elasticsearch~=7.13.1"},
    "hive": {"pyhive~=0.6.3", "thrift~=0.13.0", "sasl==0.3.1", "thrift-sasl==0.4.3"},
    "kafka": {"confluent_kafka>=1.5.0", "fastavro>=1.2.0"},
    "ldap-users": {"ldap3==2.9.1"},
    "looker": {"looker-sdk==21.12.2"},
    "mssql": {"sqlalchemy-pytds>=0.3"},
    "mssql-odbc": {"pyodbc"},
    "mysql": {"pymysql>=1.0.2"},
    "oracle": {"cx_Oracle"},
    "pii-processor": pii_requirements,
    "presto": {"pyhive~=0.6.3"},
    "postgres": {"pymysql>=1.0.2", "psycopg2-binary", "GeoAlchemy2"},
    "redshift": {"sqlalchemy-redshift", "GeoAlchemy2", "psycopg2-binary"},
    "redshift-usage": {"sqlalchemy-redshift", "psycopg2-binary", "GeoAlchemy2"},
    "scheduler": scheduler_requirements,
    "data-profiler": {"openmetadata-data-profiler"},
    "snowflake": {"snowflake-sqlalchemy<=1.2.4"},
    "snowflake-usage": {"snowflake-sqlalchemy<=1.2.4"},
    "sample-data": {"faker~=8.1.1","pandas~=1.3.1"},
    "superset": {},
    "tableau": {"tableau-api-lib==0.1.22"},
    "vertica": {"sqlalchemy-vertica[vertica-python]>=0.0.5"}
}

build_options = {"includes": ["_cffi_backend"]}
setup(
    name="openmetadata-ingestion",
    version="0.3.2",
    url="https://open-metadata.org/",
    author="OpenMetadata Committers",
    license="Apache License 2.0",
    description="Ingestion Framework for OpenMetadata",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    python_requires=">=3.8",
    options={"build_exe": build_options},
    package_dir={"": "src"},
    zip_safe=False,
    dependency_links=[

    ],
    project_urls={
        "Documentation": "https://docs.open-metadata.org/",
        "Source": "https://github.com/open-metadata/OpenMetadata",
    },
    packages=find_namespace_packages(where='./src', exclude=['tests*']),
    entry_points={
        "console_scripts": ["metadata = metadata.cmd:metadata"],
    },
    install_requires=list(base_requirements),
    extras_require={
        "base": list(base_requirements),
        **{
            plugin: list(dependencies)
            for (plugin, dependencies) in plugins.items()
        },
        "all": list(
            base_requirements.union(
                *[
                    requirements
                    for plugin, requirements in plugins.items()
                ]
            )
        )
    }

)



# generated by datamodel-codegen:
#   filename:  schema/api/services/updateMessagingService.json
#   timestamp: 2021-09-13T04:07:21+00:00

from __future__ import annotations

from typing import Optional

from pydantic import AnyUrl, BaseModel, Field

from ...entity.services import messagingService
from ...type import schedule


class UpdateMessagingServiceEntityRequest(BaseModel):
    description: Optional[str] = Field(
        None, description='Description of Messaging service entity.'
    )
    brokers: Optional[messagingService.Brokers] = None
    schemaRegistry: Optional[AnyUrl] = Field(None, description='Schema registry URL.')
    ingestionSchedule: Optional[schedule.Schedule] = Field(
        None, description='Schedule for running metadata ingestion jobs'
    )



# generated by datamodel-codegen:
#   filename:  schema/entity/bots.json
#   timestamp: 2021-09-13T04:07:21+00:00

from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, Field, constr

from ..type import basic


class Bot(BaseModel):
    id: Optional[basic.Uuid] = Field(
        None, description='Unique identifier of a bot instance.'
    )
    name: Optional[constr(min_length=1, max_length=64)] = Field(
        None, description='Name of the bot.'
    )
    displayName: Optional[str] = Field(
        None,
        description="Name used for display purposes. Example 'FirstName LastName'.",
    )
    description: Optional[str] = Field(None, description='Description of the bot.')
    href: Optional[basic.Href] = Field(
        None, description='Link to the resource corresponding to this bot.'
    )



# generated by datamodel-codegen:
#   filename:  schema/entity/services/databaseService.json
#   timestamp: 2021-09-13T04:07:21+00:00

from __future__ import annotations

from enum import Enum
from typing import Optional

from pydantic import BaseModel, Field, constr

from ...type import basic, jdbcConnection, schedule


class DatabaseServiceType(Enum):
    BigQuery = 'BigQuery'
    MySQL = 'MySQL'
    Redshift = 'Redshift'
    Snowflake = 'Snowflake'
    Postgres = 'Postgres'
    MSSQL = 'MSSQL'
    Hive = 'Hive'
    Oracle = 'Oracle'
    Athena = 'Athena'
    Presto = 'Presto'
    Vertica = 'Vertica'


class DatabaseService(BaseModel):
    id: basic.Uuid = Field(
        ..., description='Unique identifier of this database service instance.'
    )
    name: constr(min_length=1, max_length=64) = Field(
        ..., description='Name that identifies this database service.'
    )
    serviceType: DatabaseServiceType = Field(
        ...,
        description='Type of database service such as MySQL, BigQuery, Snowflake, Redshift, Postgres...',
    )
    description: Optional[str] = Field(
        None, description='Description of a database service instance.'
    )
    href: basic.Href = Field(
        ..., description='Link to the resource corresponding to this database service.'
    )
    jdbc: jdbcConnection.JdbcInfo = Field(
        ..., description='JDBC connection information.'
    )
    ingestionSchedule: Optional[schedule.Schedule] = Field(
        None, description='Schedule for running metadata ingestion jobs.'
    )



# generated by datamodel-codegen:
#   filename:  schema/entity/teams/team.json
#   timestamp: 2021-09-13T04:07:21+00:00

from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, Field, constr

from ...type import basic, entityReference, profile


class TeamName(BaseModel):
    __root__: constr(min_length=1, max_length=64) = Field(
        ...,
        description='A unique name of the team typically the team ID from an identity provider. Example - group Id from LDAP.',
    )


class Team(BaseModel):
    id: basic.Uuid
    name: TeamName
    displayName: Optional[str] = Field(
        None, description="Name used for display purposes. Example 'Data Science team'."
    )
    description: Optional[str] = Field(None, description='Description of the team.')
    href: basic.Href = Field(
        ..., description='Link to the resource corresponding to this entity.'
    )
    profile: Optional[profile.Profile] = Field(
        None, description='Team profile information.'
    )
    deleted: Optional[bool] = Field(
        None, description='When true the team has been deleted.'
    )
    users: Optional[entityReference.EntityReferenceList] = Field(
        None, description='Users that are part of the team.'
    )
    owns: Optional[entityReference.EntityReferenceList] = Field(
        None, description='List of entities owned by the team.'
    )



#  Licensed to the Apache Software Foundation (ASF) under one or more
#  contributor license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright ownership.
#  The ASF licenses this file to You under the Apache License, Version 2.0
#  (the "License"); you may not use this file except in compliance with
#  the License. You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import textwrap

TABLE_ELASTICSEARCH_INDEX_MAPPING = textwrap.dedent(
    """
    {
    "mappings":{
          "properties": {
            "table_name": {
              "type":"text",
              "analyzer": "keyword"
            },
            "schema": {
              "type":"text",
              "analyzer": "simple",
              "fields": {
                "raw": {
                  "type": "keyword"
                }
              }
            },
            "display_name": {
              "type": "keyword"
            },
            "owner": {
              "type": "keyword"
            },
            "followers": {
              "type": "keyword"
            },
            "last_updated_timestamp": {
              "type": "date",
              "format": "epoch_second"
            },
            "description": {
              "type": "text"
            },
            "tier": {
              "type": "keyword"
            },
            "column_names": {
              "type":"keyword"
            },
            "column_descriptions": {
              "type": "text"
            },
            "tags": {
              "type": "keyword"
            },
            "badges": {
              "type": "text"
            },
            "service": {
              "type": "keyword"
            },
            "service_type": {
              "type": "keyword"
            },
            "database": {
              "type": "keyword"
            },
            "suggest": {
              "type": "completion"
            },
            "monthly_stats":{
              "type": "long"
            },
            "monthly_percentile_rank":{
              "type": "long"
            },
            "weekly_stats":{
              "type": "long"
            },
            "weekly_percentile_rank":{
              "type": "long"
            },
            "daily_percentile_rank": {
             "type": "long"
            },
            "daily_stats": {
              "type": "long"
            }
          }
        }
    }
    """
)

TOPIC_ELASTICSEARCH_INDEX_MAPPING = textwrap.dedent(
    """
    {
    "mappings":{
          "properties": {
            "topic_name": {
              "type":"text",
              "analyzer": "keyword"
            },
            "schema": {
              "type":"text",
              "analyzer": "simple",
              "fields": {
                "raw": {
                  "type": "keyword"
                }
              }
            },
            "display_name": {
              "type": "keyword"
            },
            "owner": {
              "type": "keyword"
            },
            "followers": {
              "type": "keyword"
            },
            "last_updated_timestamp": {
              "type": "date",
              "format": "epoch_second"
            },
            "description": {
              "type": "text"
            },
            "tier": {
              "type": "keyword"
            },
            "tags": {
              "type": "keyword"
            },
            "service": {
              "type": "keyword"
            },
            "service_type": {
              "type": "keyword"
            },
            "suggest": {
              "type": "completion"
            }
          }
        }
    }
    """
)

DASHBOARD_ELASTICSEARCH_INDEX_MAPPING = textwrap.dedent(
    """
    {
    "mappings":{
          "properties": {
            "dashboard_name": {
              "type":"text",
              "analyzer": "keyword"
            },
            "display_name": {
              "type": "keyword"
            },
            "owner": {
              "type": "keyword"
            },
            "followers": {
              "type": "keyword"
            },
            "last_updated_timestamp": {
              "type": "date",
              "format": "epoch_second"
            },
            "description": {
              "type": "text"
            },
            "chart_names": {
              "type":"keyword"
            },
            "chart_descriptions": {
              "type": "text"
            },
            "tier": {
              "type": "keyword"
            },
            "tags": {
              "type": "keyword"
            },
            "service": {
              "type": "keyword"
            },
            "service_type": {
              "type": "keyword"
            },
            "suggest": {
              "type": "completion"
            },
             "monthly_stats":{
              "type": "long"
            },
            "monthly_percentile_rank":{
              "type": "long"
            },
            "weekly_stats":{
              "type": "long"
            },
            "weekly_percentile_rank":{
              "type": "long"
            },
            "daily_percentile_rank": {
             "type": "long"
            },
            "daily_stats": {
              "type": "long"
            }
          }
        }
    }
    """
)


#  Licensed to the Apache Software Foundation (ASF) under one or more
#  contributor license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright ownership.
#  The ASF licenses this file to You under the Apache License, Version 2.0
#  (the "License"); you may not use this file except in compliance with
#  the License. You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import pathlib

from metadata.config.common import ConfigModel
from metadata.ingestion.api.common import WorkflowContext, Record
from metadata.ingestion.api.sink import Sink, SinkStatus
from metadata.ingestion.ometa.openmetadata_rest import MetadataServerConfig

logger = logging.getLogger(__name__)


class FileSinkConfig(ConfigModel):
    filename: str


class FileSink(Sink):
    config: FileSinkConfig
    report: SinkStatus

    def __init__(self, ctx: WorkflowContext, config: FileSinkConfig, metadata_config: MetadataServerConfig):
        super().__init__(ctx)
        self.config = config
        self.metadata_config = metadata_config
        self.report = SinkStatus()

        fpath = pathlib.Path(self.config.filename)
        self.file = fpath.open("w")
        self.file.write("[\n")
        self.wrote_something = False

    @classmethod
    def create(cls, config_dict: dict, metadata_config_dict: dict, ctx: WorkflowContext):
        config = FileSinkConfig.parse_obj(config_dict)
        metadata_config = MetadataServerConfig.parse_obj(metadata_config_dict)
        return cls(ctx, config, metadata_config)

    def write_record(
        self,
        record: Record
    ) -> None:

        if self.wrote_something:
            self.file.write(",\n")

        self.file.write(record.to_json())
        self.wrote_something = True
        self.report.records_written(record)

    def get_status(self):
        return self.report

    def close(self):
        self.file.write("\n]")
        self.file.close()
